package logging

import (
	"github.com/go-chi/chi/v5/middleware"
	"gitlab.schukai.com/oss/utilities/conan/header"
	"gitlab.schukai.com/oss/utilities/conan/release"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
	"net/http"
	"time"
)

var logger *zap.Logger
var sugaredLogger *zap.SugaredLogger

func InitLogger() {

	var cfg zap.Config

	if release.GetVersion() == "dev" {
		cfg.Development = true
	} else {
		cfg.Development = false
	}

	cfg.Encoding = "console"
	cfg.OutputPaths = []string{"stdout"}
	cfg.ErrorOutputPaths = []string{"stderr"}
	cfg.EncoderConfig = zap.NewDevelopmentEncoderConfig()

	l := zapcore.Level(10)
	l = zap.InfoLevel

	switch l {
	// DebugLevel logs are typically voluminous, and are usually disabled in
	// production.
	case zap.DebugLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.DebugLevel)
	// InfoLevel is the default logging priority.
	case zap.InfoLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.InfoLevel)
	// WarnLevel logs are more important than Info, but don't need individual
	// human review.
	case zap.WarnLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.WarnLevel)
	// ErrorLevel logs are high-priority. If an application is running smoothly,
	// it shouldn't generate any error-level logs.
	case zap.ErrorLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.ErrorLevel)
	// DPanicLevel logs are particularly important errors. In development the
	// logger panics after writing the message.
	case zap.DPanicLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.DPanicLevel)
	// PanicLevel logs a message, then panics.
	case zap.PanicLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.PanicLevel)
	// FatalLevel logs a message, then calls os.Exit(1).
	case zap.FatalLevel:
		cfg.Level = zap.NewAtomicLevelAt(zap.FatalLevel)
	default:
		cfg.Level = zap.NewAtomicLevelAt(zap.ErrorLevel)

	}

	var err error
	logger, err = cfg.Build()
	if err != nil {
		panic(err)
	}
	defer logger.Sync()

	var mode string

	sugaredLogger = logger.Sugar()
	sugaredLogger.Infof("init %s logger", mode)

}

func LogInfo(template string, args ...interface{}) {
	if sugaredLogger == nil {
		InitLogger()
	}
	sugaredLogger.Infof(template, args)
}

func LogDebug(template string, args ...interface{}) {
	if sugaredLogger == nil {
		InitLogger()
	}
	sugaredLogger.Debugf(template, args)
}

func LogError(template string, args ...interface{}) {
	if sugaredLogger == nil {
		InitLogger()
	}
	sugaredLogger.Errorf(template, args)
}

func LoggerMiddleware(next http.Handler) http.Handler {

	fn := func(w http.ResponseWriter, r *http.Request) {
		start := time.Now()
		ww := middleware.NewWrapResponseWriter(w, r.ProtoMajor)
		next.ServeHTTP(ww, r)

		ctx := r.Context()
		requestIDString := ""
		requestID, ok := ctx.Value(header.XConanRequestID).(header.RequestUUID)
		if ok {
			requestIDString = requestID.String()
		}

		latency := time.Since(start)

		fields := []zapcore.Field{
			zap.Int("status", ww.Status()),
			zap.String("request-id", requestIDString),
			zap.Duration("took", latency),
			zap.Int64("measure.latency", latency.Nanoseconds()),
			zap.String("remote", r.RemoteAddr),
			zap.String("request", r.RequestURI),
			zap.String("method", r.Method),
		}

		if logger == nil {
			InitLogger()
		}
		logger.Info("request completed", fields...)

	}

	return http.HandlerFunc(fn)
}
