package watch

import (
	"bytes"
	"github.com/Masterminds/sprig/v3"
	"github.com/bitfield/script"
	"github.com/fsnotify/fsnotify"
	"gitlab.schukai.com/oss/utilities/conan/configuration"
	"gitlab.schukai.com/oss/utilities/conan/constants"
	"gitlab.schukai.com/oss/utilities/conan/logging"
	"gitlab.schukai.com/oss/utilities/conan/websocket"
	"os"
	"path"
	"path/filepath"
	"regexp"
	"strings"
	"sync"
	"text/template"
)

var (
	watcher        *fsnotify.Watcher
	watchList      map[string]string
	watchListMutex = &sync.Mutex{}
)

func init() {
	var err error

	watcher, err = fsnotify.NewWatcher()
	if err != nil {
		logging.LogError("watching: watch error: %v", err.Error())
	}
	watchList = make(map[string]string)
}

func executeWatchAction(watchPath string) {

	watch := configuration.GetWatch()

	for _, w := range watch {

		if !strings.HasPrefix(watchPath, w.Path) {
			continue
		}

		if w.Exclude != nil {

			stop := false

			for _, e := range w.Exclude {

				reg, err := regexp.Compile(e)
				if err != nil {
					logging.LogError("watching: compile exclude regex error: %v", err.Error())
					stop = true
					break
				}

				if reg.MatchString(watchPath) {
					stop = true
					break
				}
			}
			if stop {
				continue
			}
		}

		if w.Command == "" {
			logging.LogDebug("watching: no command defined for %s", watchPath)
			continue
		}

		if w.Command == "!reload" {
			websocket.SendReloadMessage()
			continue
		}

		logging.LogDebug("watching: execute watch action: %s", w.Command)

		t, err := template.New("").Funcs(sprig.FuncMap()).Parse(w.Command)
		if err != nil {
			logging.LogError("watching: execute watch action error: %v", err.Error())
			continue
		}

		data := make(map[string]interface{})
		data["Path"] = watchPath
		data["WebPath"] = configuration.GetServerWebPath()
		data["ConfigPath"] = configuration.GetConfigurationPath()
		data["PID"] = os.Getpid()
		data["Bin"] = os.Args[0]
		data["Call"] = strings.Join(os.Args[0:], " ")

		stat, err := os.Stat(watchPath)
		if err != nil {
			logging.LogError("watching: execute watch action error: %v", err.Error())
			continue
		}

		if stat.IsDir() {
			data["IsDir"] = true
			data["Directory"] = watchPath
		} else {
			data["IsDir"] = false
			data["Directory"] = path.Dir(watchPath)
		}

		var cmd bytes.Buffer
		err = t.Execute(&cmd, data)
		if err != nil {
			logging.LogError("watching: execute watch action error: %v", err.Error())
			continue
		}

		logging.LogInfo("watching: execute watch action: %s", cmd.String())

		result := script.Exec(cmd.String())
		if result.Error() != nil {
			logging.LogError("watching: execute watch action error: %v", result.Error())
		}

		r, err := result.String()
		logging.LogInfo("%s", r)

		if err != nil {
			logging.LogError("watching: execute watch action error: %v", err.Error())
			continue
		}

	}

}

func StartWatching() error {

	webPath := configuration.GetServerWebPath()

	go func() {
		for {
			select {
			case event, ok := <-watcher.Events:
				if !ok {
					return
				}

				fullPath := event.Name
				d := strings.TrimPrefix(fullPath, webPath)

				if event.Op&fsnotify.Chmod == fsnotify.Chmod {
					logging.LogDebug("watching: chmod file: %s", d)
					//RemoveFileHash(fullPath)
				}

				if event.Op&fsnotify.Write == fsnotify.Write {
					logging.LogDebug("watching: modified file: %s", d)

					if fullPath == path.Join(configuration.GetConfigurationPath(), constants.ConfigFileName) {
						configuration.Reload()
						continue
					}

					//RemoveFileHash(fullPath)
					executeWatchAction(fullPath)
				}

				if event.Op&fsnotify.Create == fsnotify.Create {
					logging.LogDebug("watching: created file: %s", d)
					scanPath(fullPath)
					executeWatchAction(fullPath)
				}

				if event.Op&fsnotify.Remove == fsnotify.Remove {
					logging.LogDebug("watching: removed file: %s", d)
					removeFromWatchlist(fullPath)
					//RemoveFileHash(fullPath)
				}

				if event.Op&fsnotify.Rename == fsnotify.Rename {
					logging.LogDebug("watching: renamed file: %s", event.Name)
					removeFromWatchlist(fullPath)
					//RemoveFileHash(fullPath)
				}

			case err, ok := <-watcher.Errors:
				if !ok {
					return
				}
				logging.LogError("watching: watch error: %v", err.Error())

			}

		}
	}()

	return nil

}

func InitWatch() {

	watchList := configuration.GetWatch()
	if len(watchList) == 0 {
		watchList = append(watchList, configuration.Watch{
			Path:    configuration.GetServerWebPath(),
			Command: "!reload",
		})
	}

	c := path.Join(configuration.GetConfigurationPath(), constants.ConfigFileName)

	if _, err := os.Stat(c); err == nil {
		watchList = append(watchList, configuration.Watch{
			Path: path.Join(configuration.GetConfigurationPath(), constants.ConfigFileName),
		})
	}

	for _, w := range watchList {
		scanPath(w.Path)
	}

}

func scanPath(p string) {

	if !path.IsAbs(p) {
		p = path.Join(configuration.GetConfigurationPath(), p)
	}

	followFlag := configuration.GetServerFlagFollowSymlinks()

	err := filepath.Walk(p,
		func(x string, info os.FileInfo, err error) error {
			if err != nil {
				return err
			}

			if info.IsDir() && p == x {
				addToWatchlist(x)
				return nil
			}

			if followFlag {
				x2, err := filepath.EvalSymlinks(x)
				if err != nil {
					return err
				}

				if x2 != x {
					scanPath(x2)
					return nil
				}
			}

			addToWatchlist(x)

			return nil
		})

	if err != nil {
		logging.LogError("watching: watch files error %s", err.Error())
	}
}

func removeFromWatchlist(p string) {

	watchListMutex.Lock()
	for k, _ := range watchList {
		if k == p {
			delete(watchList, k)
			break
		}
	}

	watchListMutex.Unlock()
}

func addToWatchlist(p string) {

	watchListMutex.Lock()
	err := watcher.Add(p)
	if err != nil {
		logging.LogError("watching: watch error: %v", err.Error())
	} else {
		logging.LogDebug("watching: watching %s", p)
		watchList[p] = p
	}

	watchListMutex.Unlock()

}

func DestroyApps() {
	watcher.Close()
}
